import os
import shutil
from typing import Any, Tuple

import py
from tox import action, config, hookimpl, reporter, session, venv
from tox.package.view import create_session_view
from tox.util.lock import hold_lock
from tox.util.path import ensure_empty_dir

from .utils import (
    clone_pdm_files,
    get_env_lib_path,
    inject_pdm_to_commands,
    set_default_kwargs,
)


@hookimpl
def tox_addoption(parser: config.Parser) -> Any:
    parser.add_testenv_attribute(
        "sections", "line-list", "Specify the dependency sections to install"
    )
    os.environ["TOX_TESTENV_PASSENV"] = "PYTHONPATH"
    parser.add_argument("--pdm", default="pdm", help="The executable path of PDM")
    set_default_kwargs(venv.VirtualEnv._pcall, venv=False)
    set_default_kwargs(venv.VirtualEnv.getcommandpath, venv=False)


@hookimpl
def tox_testenv_create(venv: venv.VirtualEnv, action: action.Action) -> Any:
    clone_pdm_files(venv)

    config_interpreter = venv.getsupportedinterpreter()
    venv._pcall(
        [venv.envconfig.config.option.pdm, "use", "-f", config_interpreter],
        cwd=venv.path,
        venv=False,
        action=action,
    )
    return True


def get_package(
    session: session.Session, venv: venv.VirtualEnv
) -> Tuple[py.path.local, py.path.local]:
    config = session.config
    if config.skipsdist:
        reporter.info("skipping sdist step")
        return None
    lock_file = session.config.toxworkdir.join(
        "{}.lock".format(session.config.isolated_build_env)
    )

    with hold_lock(lock_file, reporter.verbosity0):
        package = acquire_package(config, venv)
        session_package = create_session_view(package, config.temp_dir)
        return session_package, package


def acquire_package(config: config.Config, venv: venv.VirtualEnv) -> py.path.local:
    target_dir: py.path.local = config.toxworkdir.join(config.isolated_build_env)
    ensure_empty_dir(target_dir)
    args = [venv.envconfig.config.option.pdm, "build", "--no-wheel", "-d", target_dir]
    with venv.new_action("buildpkg") as action:
        venv._pcall(
            args, cwd=venv.envconfig.config.toxinidir, venv=False, action=action
        )
        path = next(target_dir.visit("*.tar.gz"))
        action.setactivity("buildpkg", path)
        return path


@hookimpl
def tox_package(session: session.Session, venv: venv.VirtualEnv) -> Any:
    if not hasattr(session, "package"):
        session.package, session.dist = get_package(session, venv)
    # Patch the install command to install to local __pypackages__ folder
    for i, arg in enumerate(venv.envconfig.install_command):
        if arg == "python":
            venv.envconfig.install_command[i] = venv.getsupportedinterpreter()
    venv.envconfig.install_command.extend(["-t", get_env_lib_path(venv)])
    return session.package


@hookimpl
def tox_testenv_install_deps(venv: venv.VirtualEnv, action: action.Action) -> Any:
    clone_pdm_files(venv)
    sections = venv.envconfig.sections or []
    if not venv.envconfig.skip_install or sections:
        action.setactivity("pdminstall", sections)
        args = [venv.envconfig.config.option.pdm, "install", "-p", str(venv.path)]
        if "default" in sections:
            sections.remove("default")
        elif venv.envconfig.skip_install:
            args.append("--no-default")
        for section in sections:
            args.extend(["-s", section])
        venv._pcall(
            args,
            cwd=venv.envconfig.config.toxinidir,
            venv=False,
            action=action,
        )

    deps = venv.get_resolved_dependencies()
    if deps:
        depinfo = ", ".join(map(str, deps))
        action.setactivity("installdeps", depinfo)
        venv._install(deps, action=action)

    lib_path = get_env_lib_path(venv)
    bin_dir = lib_path.join("bin")
    scripts_dir = lib_path.dirpath("Scripts" if os.name == "nt" else "bin")
    print(scripts_dir)
    if os.path.exists(bin_dir):
        for item in os.listdir(bin_dir):
            bin_item = os.path.join(bin_dir, item)
            shutil.move(bin_item, os.path.join(scripts_dir, item))
    return True


@hookimpl
def tox_runtest_pre(venv: venv.VirtualEnv) -> Any:
    inject_pdm_to_commands(venv, venv.envconfig.commands_pre)
    inject_pdm_to_commands(venv, venv.envconfig.commands)
    inject_pdm_to_commands(venv, venv.envconfig.commands_post)


@hookimpl
def tox_runenvreport(venv: venv.VirtualEnv, action: action.Action):
    command = venv.envconfig.list_dependencies_command
    for i, arg in enumerate(command):
        if arg == "python":
            command[i] = venv.getsupportedinterpreter()
    venv.envconfig.list_dependencies_command.extend(["--path", get_env_lib_path(venv)])



import functools
import inspect
import os
import subprocess
from typing import List

import py
import toml
from tox import venv


def clone_pdm_files(venv: venv.VirtualEnv) -> None:
    """Initialize the PDM project for the given VirtualEnv by cloning
    the pyproject.toml and pdm.lock files to the venv path.
    """
    venv.path.ensure(dir=1)
    project_root: py.path.local = venv.envconfig.config.toxinidir
    old_pyproject = toml.loads(project_root.join("pyproject.toml").read())
    if "name" in old_pyproject.get("project", {}):
        del old_pyproject["project"]["name"]
    with venv.path.join("pyproject.toml").open("w", ensure=True) as f:
        toml.dump(old_pyproject, f)

    if project_root.join("pdm.lock").exists():
        project_root.join("pdm.lock").copy(venv.path.join("pdm.lock"))

    base_path = venv.path.join(
        f"__pypackages__/{get_version_bit(venv.getsupportedinterpreter())}"
    )
    for subdir in ("include", "lib", "Scripts" if os.name == "nt" else "bin", "src"):
        base_path.join(subdir).ensure(dir=1)


def set_default_kwargs(func_or_method, **kwargs):
    """Change the default value for keyword arguments."""
    func = getattr(func_or_method, "__func__", func_or_method)
    args = inspect.signature(func).parameters
    defaults = {k: v.default for k, v in args.items() if v.default is not v.empty}
    for k, v in kwargs.items():
        if k in defaults:
            defaults[k] = v
    func.__defaults__ = tuple(defaults.values())


@functools.lru_cache()
def get_version_bit(executable: os.PathLike) -> str:
    """Get the version of the Python interperter.

    :param executable: The path of the python executable
    :param as_string: return the version string if set to True
        and version tuple otherwise
    :param digits: the number of version parts to be returned
    :returns: A tuple of (version, is_64bit)
    """
    script = (
        'import os,sys;print(".".join(map(str, sys.version_info[:2])) '
        '+ ("-32" if os.name == "nt" and sys.maxsize <= 2**32 else ""))'
    )
    args = [executable, "-c", script]
    return subprocess.check_output(args).decode().strip()


def get_env_lib_path(venv: venv.VirtualEnv) -> py.path.local:
    """Return the PEP 582 library path for the given VirtualEnv."""
    version_bit = get_version_bit(venv.getsupportedinterpreter())
    return venv.path.join(f"__pypackages__/{version_bit}/lib")


NON_PROJECT_COMMANDS = ("cache", "show", "completion")


def inject_pdm_to_commands(venv: venv.VirtualEnv, commands: List[List[str]]) -> None:
    """Inject pdm run to the commands in place.

    Examples:
        - <cmd> -> pdm run <cmd>
        - pip install <pkg> -> pdm run pip install <pkg> -t <lib_path>
    """
    cmd_prefix = [venv.envconfig.config.option.pdm, "run", "-p", venv.path]
    pip_postfix = ["-t", get_env_lib_path(venv)]

    for argv in commands:
        cmd = argv[0]
        inject_pos = 0
        if cmd == "-":
            cmd = argv[1]
            inject_pos = 1
        elif cmd.startswith("-"):
            cmd = cmd.lstrip("-")
            argv[:1] = ["-", cmd]
            inject_pos = 1
        if argv[inject_pos : inject_pos + 2] == ["pip", "install"] or argv[
            inject_pos : inject_pos + 4
        ] == ["python", "-m", "pip", "install"]:
            argv.extend(pip_postfix)
        if argv[inject_pos : inject_pos + 3] == ["python", "-m", "pdm"]:
            argv[inject_pos : inject_pos + 3] = ["pdm"]
        if argv[inject_pos] == "pdm":
            argv[inject_pos] = venv.envconfig.config.option.pdm
            if (
                inject_pos + 1 < len(argv)
                and argv[inject_pos + 1] not in NON_PROJECT_COMMANDS
                and not argv[inject_pos + 1].startswith("-")
            ):
                argv[inject_pos + 2 : inject_pos + 2] = ["-p", venv.path]
        else:
            argv[inject_pos:inject_pos] = cmd_prefix
